# Copyright 2017 Andreas Kirsch <blackhc@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Linear programming solver for MDPs.

This is a very basic solver.
"""

import numpy as np

from blackhc import mdp


class LinearProgramming(object):
    def __init__(self, mdp_spec: mdp.MDPSpec):
        self.discount = mdp_spec.discount
        self.num_states = mdp_spec.num_states
        self.num_actions = mdp_spec.num_actions
        self.mdp_spec = mdp_spec

        transitions = mdp.Transitions(mdp_spec)
        next_states = np.zeros(shape=(self.num_states, self.num_actions, self.num_states))
        expected_rewards = np.zeros(shape=(self.num_states, self.num_actions))
        for (state, action), choices in transitions.next_states.items():
            for next_state, prob in choices.items():
                next_states[state.index, action.index, next_state.index] = prob
        for state in mdp_spec.states:
            if state.terminal_state:
                next_states[state.index, :, state.index] = 1.

        for (state, action), choices in transitions.rewards.items():
            expected_rewards[state.index, action.index] = sum(value * prob for value, prob in choices.items())

        self.next_states = next_states
        self.expected_rewards = expected_rewards

    def compute_q_table(self, max_iterations=100, all_close=None):
        return _fix_point_iterate(self.expected_rewards.copy(),
                                  lambda q_table: self.q_table_from_v_vector(self.v_vector_from_q_table(q_table)),
                                  max_iterations=max_iterations,
                                  all_close=all_close
                                  )

    def compute_v_vector(self, max_iterations=100, all_close=None):
        return _fix_point_iterate(np.zeros((self.num_states,)),
                                  lambda v_vector: self.v_vector_from_q_table(self.q_table_from_v_vector(v_vector)),
                                  max_iterations=max_iterations,
                                  all_close=all_close)

    # noinspection PyMethodMayBeStatic
    def v_vector_from_q_table(self, q_table):
        v_vector = q_table.max(axis=-1)
        return v_vector

    def q_table_from_v_vector(self, v_vector):
        # TODO: somebody said that q table converges faster than v table
        # Is that true? What if take computation cost into account?
        # TODO: support passing it in as a parameter
        q_table = self.expected_rewards + self.discount * (
            self.next_states * v_vector.reshape((1, 1, self.num_states))).sum(axis=-1)
        return q_table


def _fix_point_iterate(initial, iterate, max_iterations, all_close=None):
    if not all_close:
        all_close = np.allclose

    value = initial
    for _ in range(max_iterations):
        next_value = iterate(value)
        if all_close(value, next_value):
            return next_value
        value = next_value
    raise ValueError('No convergence after %s iterations!\n%s' % (max_iterations, value))



# Copyright 2017 Andreas Kirsch <blackhc@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib

from blackhc import mdp
from blackhc.mdp.dsl import context as dsl_context
from blackhc.mdp.dsl import ast


# noinspection PyShadowingBuiltins
SyntaxError = ast.DslSyntaxError


def state(name=None):
    new_state = dsl_context.mdp_spec.state(name)
    return ast.State(new_state)


def terminal_state(name=None):
    new_state = dsl_context.mdp_spec.state(name, terminal_state=True)
    return ast.State(new_state)


def action(name=None):
    new_action = dsl_context.mdp_spec.action(name)
    return ast.Action(new_action)


def reward(value):
    return ast.Reward(mdp.Reward(value))


def to_env():
    return dsl_context.mdp_spec.to_env()


def to_graph(*args, **kwargs):
    return dsl_context.mdp_spec.to_graph(*args, **kwargs)

def discount(value):
    dsl_context.mdp_spec.discount = value


@contextlib.contextmanager
def new():
    old_context = dsl_context.mdp_spec
    dsl_context.mdp_spec = mdp.MDPSpec()
    yield dsl_context.mdp_spec
    dsl_context.mdp_spec = old_context



# Copyright 2017 Andreas Kirsch <blackhc@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing

from blackhc import mdp
from blackhc.mdp.dsl import context


class DslSyntaxError(SyntaxError):
    pass


class TransitionInfo(object):
    """Whether an expression has a state, action or outcome set."""

    def __init__(self, has_action=False, has_state=False, has_outcome=False):
        # Whether there is an action column
        self.has_action = has_action
        # Whether there is a state column
        self.has_state = has_state
        # Whether there is an outcome column
        self.has_outcome = has_outcome

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__dict__)

    @property
    def fully_specified(self):
        """Whether all three attributes have been provided and transitions are thus fully specified."""
        return self.has_state and self.has_action and self.has_outcome


class Node(object):
    def __init__(self):
        self._transition_info = None

    @property
    def transition_info(self) -> TransitionInfo:
        if not self._transition_info:
            self._transition_info = self.apply(TransitionInfoVisitor())
        return self._transition_info

    def apply(self, visitor: 'NodeVisitor') -> typing.Any:
        return visitor.visit_atom(self)
    
    def __or__(self, other):
        return Alternatives([self]) | other

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__dict__)


class SupportConjunction(object):
    def __and__(self: Node, right):
        return Conjunction(self, right)


class SupportMapping(object):
    def __gt__(self: Node, outcome):
        return Mapping(self, outcome)


class Transition(object):
    """Immutable transition call (by gentleman's agreement)."""

    def __init__(self, state=None, action=None, outcome=None):
        self.state = state
        self.action = action
        self.outcome = outcome

    def replace(self, state=None, action=None, outcome=None):
        def not_already_set(new, current):
            return not new or not current
        # TODO: replace with value errors!
        assert not_already_set(state, self.state), (state, self.state)
        assert not_already_set(action, self.action), (action, self.action)
        assert not_already_set(outcome, self.outcome), (outcome, self.outcome)

        state = state or self.state
        action = action or self.action
        outcome = outcome or self.outcome
        return Transition(state, action, outcome)


class Alternatives(Node, SupportConjunction, SupportMapping):
    def __init__(self, alternatives: typing.List[Node]):
        super().__init__()

        self.alternatives = alternatives

    def apply(self, visitor: 'NodeVisitor'):
        return visitor.visit_alternatives(self)

    def __or__(self, other: Node):
        if isinstance(other, Alternatives):
            # Merging alternatives into a single one.
            return Alternatives(self.alternatives + other.alternatives)
        else:
            return Alternatives(self.alternatives + [other])

    def __mul__(self, prob):
        # noinspection PyUnresolvedReferences
        return Alternatives([alternative * prob for alternative in self.alternatives])
    

class Reward(Node):
    def __init__(self, reward: mdp.Reward):
        super().__init__()

        self.reward = reward

    def __mul__(self, other):
        return Reward(mdp.Reward(self.reward.outcome, self.reward.weight * other))

    def apply(self, visitor: 'NodeVisitor'):
        return visitor.visit_reward(self)


class Action(Node, SupportConjunction, SupportMapping):
    def __init__(self, action):
        super().__init__()

        self.action = action

    def apply(self, visitor: 'NodeVisitor'):
        return visitor.visit_action(self)


class State(Node, SupportConjunction, SupportMapping):
    def __init__(self, state):
        super().__init__()

        self.state = state

    def __mul__(self, other):
        return WeightedState(mdp.NextState(self.state, other))

    def __gt__(self, other):
        return Mapping(self, other)

    def apply(self, visitor: 'NodeVisitor'):
        return visitor.visit_state(self)


class WeightedState(Node):
    def __init__(self, next_state: mdp.NextState):
        super().__init__()

        self.next_state = next_state

    def __mul__(self, other):
        return WeightedState(mdp.NextState(self.next_state.outcome, self.next_state.weight * other))

    def apply(self, visitor: 'NodeVisitor'):
        return visitor.visit_weighted_state(self)


class Conjunction(Node, SupportMapping):
    def __init__(self, left: Node, right: Node):
        super().__init__()

        left.apply(TriggerTypeVerifier())
        # right can be either a mapping (so have an outcome) or be a trigger
        if not right.transition_info.has_outcome:
            # right must be a trigger
            right.apply(TriggerTypeVerifier())
        # else: Mapping has already verified itself.

        self.left = left
        self.right = right

        if self.transition_info.fully_specified:
            compile_transitions(self)

    def apply(self, visitor: 'NodeVisitor'):
        return visitor.visit_conjunction(self)


class Mapping(Node):
    def __init__(self, trigger: Node, outcome: Node):
        super().__init__()

        trigger.apply(TriggerTypeVerifier())
        outcome.apply(OutcomeTypeVerifier())

        self.trigger = trigger
        self.outcome = outcome

        if self.transition_info.fully_specified:
            compile_transitions(self)

    def apply(self, visitor: 'NodeVisitor'):
        return visitor.visit_mapping(self)


def compile_transitions(node):
    transitions = node.apply(TransitionVisitor([Transition()]))
    for transition in transitions:
        if transition.state.terminal_state:
            raise DslSyntaxError('Attempted to specify transition %s for terminal state!' % transition)
        context.mdp_spec.transition(transition.state, transition.action, transition.outcome)


class NodeVisitor(object):
    def visit_atom(self, node: Node):
        raise AssertionError('This should be unreachable!')

    def visit_alternatives(self, node: Alternatives):
        return self.visit_atom(node)

    def visit_reward(self, node: Reward):
        return self.visit_atom(node)

    def visit_action(self, node: Action):
        return self.visit_atom(node)

    def visit_state(self, node: State):
        return self.visit_atom(node)

    def visit_weighted_state(self, node: WeightedState):
        return self.visit_atom(node)

    def visit_conjunction(self, node: Conjunction):
        return self.visit_atom(node)

    def visit_mapping(self, node: Mapping):
        return self.visit_atom(node)


class OutcomeTypeVerifier(NodeVisitor):
    """Verifies that all nodes are valid as outcomes."""

    def fail(self, node):
        raise DslSyntaxError('%s not a valid outcome of a transition (either Reward or State)!' % node)

    def visit_weighted_state(self, node: WeightedState):
        pass

    def visit_alternatives(self, node: Alternatives):
        for alternative in node.alternatives:
            alternative.apply(self)

    def visit_mapping(self, node: Mapping):
        # Cannot be valid because a mapping's left side cannot be an outcome.
        self.fail(node)

    def visit_action(self, node: Action):
        self.fail(node)

    def visit_conjunction(self, node: Conjunction):
        # A conjunction cannot be an outcome.
        self.fail(node)

    def visit_reward(self, node: Reward):
        pass

    def visit_state(self, node: State):
        pass


class TriggerTypeVerifier(NodeVisitor):
    """Verifies that all types are valid as triggers."""

    def fail(self, node):
        raise DslSyntaxError('%s not a valid trigger for a transition (either a State or an Action)!' % node)

    def visit_alternatives(self, node: Alternatives):
        for alternative in node.alternatives:
            alternative.apply(self)
        if not node.transition_info.has_action and not node.transition_info.has_state:
            raise DslSyntaxError("%s contains non-homogeneous alternatives!" % node)

    def visit_reward(self, node: Reward):
        pass

    def visit_weighted_state(self, node: WeightedState):
        self.fail(node)

    def visit_state(self, node: State):
        pass

    def visit_mapping(self, node: Mapping):
        self.fail(node)

    def visit_conjunction(self, node: Conjunction):
        node.left.apply(self)
        node.right.apply(self)

    def visit_action(self, node: Action):
        pass


class TransitionInfoVisitor(NodeVisitor):
    def visit_weighted_state(self, node: WeightedState):
        return TransitionInfo()

    def visit_alternatives(self, node: Alternatives):
        has_action = True
        has_state = True
        has_outcome = True
        for alternative in node.alternatives:
            has_state = has_state and alternative.transition_info.has_state
            has_action = has_action and alternative.transition_info.has_action
            has_outcome = has_outcome and alternative.transition_info.has_outcome
        return TransitionInfo(has_state=has_state, has_action=has_action, has_outcome=has_outcome)

    def visit_action(self, node: Action):
        return TransitionInfo(has_action=True)

    def visit_state(self, node: State):
        return TransitionInfo(has_state=True)

    def visit_reward(self, node: Reward):
        return TransitionInfo()

    def visit_conjunction(self, node: Conjunction):
        left_transition_info = node.left.transition_info
        right_transition_info = node.right.transition_info

        if left_transition_info.has_state and right_transition_info.has_state:
            raise DslSyntaxError('%s and %s both have state information!' % (node.left, node.right))
        if left_transition_info.has_action and right_transition_info.has_action:
            raise DslSyntaxError('%s and %s both have action information!' % (node.left, node.right))

        return TransitionInfo(has_action=left_transition_info.has_action or right_transition_info.has_action,
                              has_state=left_transition_info.has_state or right_transition_info.has_state,
                              has_outcome=right_transition_info.has_outcome)

    def visit_mapping(self, node: Mapping):
        return TransitionInfo(has_action=node.trigger.transition_info.has_action,
                              has_state=node.trigger.transition_info.has_state,
                              has_outcome=True)


class TransitionVisitor(NodeVisitor):
    """Returns a generator yielding all transitions.

    Assumes that the root node's transition info is fully specified."""
    def __init__(self, iterator: typing.Iterable[Transition]):
        self.iterator = iterator

    def visit_action(self, node: Action):
        for transition in self.iterator:
            yield transition.replace(action=node.action)

    def visit_state(self, node: State):
        for transition in self.iterator:
            yield transition.replace(state=node.state)

    def visit_alternatives(self, node: Alternatives):
        transitions = list(self.iterator)
        for alternative in node.alternatives:
            yield from alternative.apply(TransitionVisitor(transitions))

    def visit_mapping(self, node: Mapping):
        return node.outcome.apply(TransitionOutcomeVisitor(node.trigger.apply(self)))

    def visit_conjunction(self, node: Conjunction):
        return node.right.apply(TransitionVisitor(node.left.apply(self)))


class TransitionOutcomeVisitor(NodeVisitor):
    def __init__(self, iterator: typing.Iterable[Transition]):
        self.iterator = iterator

    def visit_alternatives(self, node: Alternatives):
        transitions = list(self.iterator)
        for alternative in node.alternatives:
            yield from alternative.apply(TransitionOutcomeVisitor(transitions))

    def visit_reward(self, node: Reward):
        for transition in self.iterator:
            yield transition.replace(outcome=node.reward)

    def visit_weighted_state(self, node: WeightedState):
        for transition in self.iterator:
            yield transition.replace(outcome=node.next_state)

    def visit_state(self, node: State):
        for transition in self.iterator:
            yield transition.replace(outcome=mdp.NextState(node.state))



# Copyright 2017 Andreas Kirsch <blackhc@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing
from collections import defaultdict

import gym
import gym.spaces
import networkx as nx
import numpy as np

from blackhc.mdp.version import VERSION as __version__


class State(object):
    def __init__(self, name, index, terminal_state=False):
        self.name = name
        self.index = index
        self.terminal_state = terminal_state

    def __repr__(self):
        return 'State(%s, %s, %s)' % (self.name, self.index, self.terminal_state)


class Action(object):
    def __init__(self, name, index):
        self.name = name
        self.index = index

    def __repr__(self):
        return 'Action(%s, %s)' % (self.name, self.index)


class Outcome(object):
    """An outcome can be either a reward or a next state.

    For a given (state, action) transition all potential outcomes
    are weighted according to their `weight` and normalized.
    """

    def __init__(self, outcome, weight):
        self.weight = weight
        self.outcome = outcome

    @staticmethod
    def get_choices(outcomes: typing.Iterable['Outcome']):
        """Normalize outcomes and deduplicate into a Dict[outcome, probability]."""

        # Deduplicate elements
        deduped_outcomes = defaultdict(float)
        total_weight = 0.
        for outcome in outcomes:
            deduped_outcomes[outcome.outcome] += outcome.weight
            total_weight += outcome.weight

        return {outcome: weight / total_weight for outcome, weight in deduped_outcomes.items()}


class Reward(Outcome):
    def __init__(self, value, weight=1.0):
        super().__init__(value, weight)

    def __repr__(self):
        return 'Reward(%s, %s)' % (self.outcome, self.weight)

    @staticmethod
    def get_choices(rewards: typing.Iterable['Reward']):
        return Outcome.get_choices(rewards) or {0.: 1.}


class NextState(Outcome):
    def __init__(self, state, weight=1.0):
        super().__init__(state, weight)

    def __repr__(self):
        return 'NextState(%s, %s)' % (self.outcome, self.weight)

    @staticmethod
    def get_choices(next_states: typing.Iterable['NextState']):
        return Outcome.get_choices(next_states)


class MDPSpec(object):
    def __init__(self):
        self._states = {}
        self._actions = {}
        self.states = []
        self.actions = []
        self.state_outcomes: typing.Dict[tuple, typing.List[NextState]] = defaultdict(list)
        self.reward_outcomes: typing.Dict[tuple, typing.List[Reward]] = defaultdict(list)
        self.discount = 1.0

    def state(self, name=None, terminal_state=False):
        if not name:
            if not terminal_state:
                name = 'S%s' % self.num_states
            else:
                name = 'T%s' % self.num_states

        if name not in self.states:
            new_state = State(name, self.num_states, terminal_state=terminal_state)
            self._states[name] = new_state
            self.states.append(new_state)
        return self._states[name]

    def action(self, name=None):
        if not name:
            name = 'A%s' % self.num_actions

        if name not in self.actions:
            new_action = Action(name, self.num_actions)
            self._actions[name] = new_action
            self.actions.append(new_action)
        return self._actions[name]

    def transition(self, state: State, action: Action, outcome: Outcome):
        """Specify either a next state or a reward as `outcome` for a transition."""

        if isinstance(outcome, NextState):
            self.state_outcomes[state, action].append(outcome)
        elif isinstance(outcome, Reward):
            self.reward_outcomes[state, action].append(outcome)
        else:
            raise NotImplementedError()

    @property
    def num_states(self):
        return len(self._states)

    @property
    def num_actions(self):
        return len(self.actions)

    @property
    def is_deterministic(self):
        for state in self.states:
            for action in self.actions:
                if len(self.reward_outcomes[state, action]) > 1:
                    return False
                if len(self.state_outcomes[state, action]) > 1:
                    return False
        return True

    def __repr__(self):
        return 'Mdp(states=%s, actions=%s, state_outcomes=%s, reward_outcomes=%s)' % (
            self.states, self.actions, dict(self.state_outcomes), dict(self.reward_outcomes))

    def to_graph(self, highlight_state: State = None, highlight_action: Action = None,
                 highlight_next_state: State = None):
        transitions = Transitions(self)

        graph = nx.MultiDiGraph()
        for state in self.states:
            graph.add_node(state, label=state.name)
            attributes = graph.node[state]
            if state.terminal_state:
                attributes['shape'] = 'doublecircle'
            if state == highlight_state:
                attributes['fillcolor'] = 'yellow'
                attributes['style'] = 'filled'
            if state == highlight_next_state:
                attributes['fillcolor'] = 'red'
                attributes['style'] = 'filled'

        for state in self.states:
            if not state.terminal_state:
                for action in self.actions:
                    reward_probs = transitions.rewards[state, action].items()
                    expected_reward = sum(reward * prob for reward, prob in reward_probs)
                    stddev_reward = (sum(
                        reward * reward * prob for reward, prob in
                        reward_probs) - expected_reward * expected_reward) ** 0.5

                    action_label = '%s %+.2f' % (action.name, expected_reward)
                    if len(reward_probs) > 1:
                        action_label += ' (%.2f)' % stddev_reward

                    next_states = transitions.next_states[state, action].items()
                    if len(next_states) > 1:
                        transition = (state, action)

                        graph.add_node(transition, shape='point')
                        graph.add_edge(state, transition,
                                       label=action_label)

                        for next_state, prob in next_states:
                            if not prob:
                                continue
                            graph.add_edge(transition, next_state, label='%3.2f%%' % (prob * 100))

                        if state == highlight_state and action == highlight_action:
                            attributes = graph.node[transition]
                            attributes['style'] = 'bold'

                            attributes = graph.edge[state][transition][0]
                            attributes['style'] = 'bold'
                            attributes['color'] = 'green'
                            if highlight_next_state:
                                # Could also check that highlight_next_state is really a next state.
                                attributes = graph.edge[transition][highlight_next_state][0]
                                attributes['style'] = 'bold'
                                attributes['color'] = 'red'
                    else:
                        next_state, _ = list(next_states)[0]
                        graph.add_edge(state, next_state, key=action,
                                       label=action_label)
                        if state == highlight_state and action == highlight_action:
                            attributes = graph.edge[state][next_state][action]
                            attributes['style'] = 'bold'
                            attributes['color'] = 'red'

        return graph

    def to_env(self):
        return MDPEnv(self)

    def validate(self):
        # For now, just validate by trying to compute the transitions.
        # It will raise errors if anything is wrong.
        Transitions(self)
        return self


class Transitions(object):
    """Container for transition probabilities."""

    def __init__(self, mdp: MDPSpec):
        self.next_states = {}
        self.rewards = {}
        for state in mdp.states:
            for action in mdp.actions:
                next_states = NextState.get_choices(mdp.state_outcomes[state, action])
                if not state.terminal_state and not next_states:
                    raise ValueError('No next states specified for non-terminal (%s, %s)!' % (state, action))
                if state.terminal_state and next_states:
                    raise ValueError('Next states %s specified for terminal (%s, %s)!' % (next_states, state, action))
                self.next_states[state, action] = next_states

                rewards = mdp.reward_outcomes[state, action]
                if state.terminal_state and rewards:
                    raise ValueError('Rewards %s specified for terminal (%s, %s)!' % (next_states, state, action))
                self.rewards[state, action] = Reward.get_choices(rewards)

    def __repr__(self):
        return 'Transitions(%s)' % self.__dict__


class MDPEnv(gym.Env):
    metadata = {'render.modes': ['human', 'rgb_array', 'png']}

    def __init__(self, mdp: MDPSpec, start_state: State = None):
        self.render_widget = None

        self.mdp = mdp
        self.transitions = Transitions(mdp)

        self._previous_state: State = None
        self._previous_action: Action = None
        self._state: State = None
        self._is_done = True
        self.observation_space = gym.spaces.Discrete(self.mdp.num_states)
        self.action_space = gym.spaces.Discrete(self.mdp.num_actions)
        self.start_state = start_state or list(self.mdp.states)[0]

    def _reset(self):
        self._previous_state = None
        self._previous_action = None
        self._state = self.start_state
        self._is_done = self._state.terminal_state
        return self._state.index

    def _step(self, action_index):
        action = self.mdp.actions[action_index]
        self._previous_state = self._state
        self._previous_action = action

        if not self._is_done:
            reward_probs = self.transitions.rewards[self._state, action]
            reward = np.random.choice(list(reward_probs.keys()), p=list(reward_probs.values()))

            next_state_probs = self.transitions.next_states[self._state, action]
            self._state = np.random.choice(list(next_state_probs.keys()), p=list(next_state_probs.values()))
            self._is_done = self._state.terminal_state
        else:
            reward = 0

        return self._state.index, reward, self._is_done, None

    def to_graph(self):
        graph = self.mdp.to_graph(highlight_state=self._previous_state, highlight_action=self._previous_action,
                                  highlight_next_state=self._state)
        return graph

    def _render(self, mode='human', close=False):
        if close:
            if self.render_widget:
                self.render_widget.close()
            return

        png_data = graph_to_png(self.to_graph())

        if mode == 'human':
            # TODO: use OpenAI's SimpleImageViewer wrapper when not running in IPython.
            if not self.render_widget:
                from IPython.display import display
                import ipywidgets as widgets

                self.render_widget = widgets.Image()
                display(self.render_widget)

            self.render_widget.value = png_data
        elif mode == 'rgb_array':
            from matplotlib import pyplot
            import io
            return pyplot.imread(io.BytesIO(png_data))
        elif mode == 'png':
            return png_data


def graph_to_png(graph):
    pydot_graph = nx.nx_pydot.to_pydot(graph)
    return pydot_graph.create_png()


def display_mdp(mdp: MDPSpec):
    from IPython.display import display, Image
    display(Image(graph_to_png(mdp.to_graph())))



# Copyright 2017 Andreas Kirsch <blackhc@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from blackhc.mdp import dsl


# noinspection PyStatementEffect
def _one_round_dmdp():
    with dsl.new() as mdp:
        start = dsl.state()
        end = dsl.terminal_state()

        action_0 = dsl.action()
        action_1 = dsl.action()

        start & (action_0 | action_1) > end
        start & action_1 > dsl.reward(1.)

        return mdp.validate()


# noinspection PyStatementEffect
def _two_round_dmdp():
    with dsl.new() as mdp:
        start = dsl.state()
        better = dsl.state()
        worse = dsl.state()
        end = dsl.terminal_state()

        action_0 = dsl.action()
        action_1 = dsl.action()

        start & action_0 > better
        better & action_1 > dsl.reward(3)

        start & action_1 > worse
        worse & action_0 > dsl.reward(1)
        worse & action_1 > dsl.reward(2)

        (better | worse) & (action_0 | action_1) > end

        return mdp.validate()


# noinspection PyStatementEffect
def _one_round_nmdp():
    with dsl.new() as mdp:
        start = dsl.state()
        end = dsl.terminal_state()

        action_0 = dsl.action()
        action_1 = dsl.action()

        start & action_0 > dsl.reward(0) | dsl.reward(5)
        start & action_1 > dsl.reward(1) | dsl.reward(3)

        start & (action_0 | action_1) > end

        return mdp.validate()


# noinspection PyStatementEffect
def _two_round_nmdp():
    with dsl.new() as mdp:
        start = dsl.state()
        a = dsl.state()
        b = dsl.state()
        end = dsl.terminal_state()

        action_0 = dsl.action()
        action_1 = dsl.action()

        start & action_0 > a
        a & action_0 > dsl.reward(-1) | dsl.reward(1)
        a & action_1 > dsl.reward(0) * 2 | dsl.reward(9)

        start & action_1 > b
        b & action_0 > dsl.reward(0) | dsl.reward(2)
        b & action_1 > dsl.reward(2) | dsl.reward(3)

        (a | b) & (action_0 | action_1) > end

        return mdp.validate()


# noinspection PyStatementEffect,PyPep8Naming
def  _multi_round_nmdp():
    with dsl.new() as mdp:
        start = dsl.state()
        end = dsl.terminal_state()

        start & dsl.action() > dsl.reward(5) | start | end * 2
        start & dsl.action() > dsl.reward(3) | start * 2 | end

        dsl.discount(0.5)

        return mdp.validate()


ONE_ROUND_DMDP = _one_round_dmdp()
TWO_ROUND_DMDP = _two_round_dmdp()

ONE_ROUND_NMDP = _one_round_nmdp()
TWO_ROUND_NMDP = _two_round_nmdp()

MULTI_ROUND_NDMP = _multi_round_nmdp()